@@ -1,6 +1,6 @@
 /*
- * Copyright 2002 Lutz Mller <lutz@users.sourceforge.net>
- * Copyright 2004-2010 Marcus Meissner <marcus@jet.franken.de>
+ * Copyright 2002 Lutz Mueller <lutz@users.sourceforge.net>
+ * Copyright 2004-2013 Marcus Meissner <marcus@jet.franken.de>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -14,8 +14,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA  02110-1301  USA
  */
 
 #include "config.h"
@@ -149,19 +149,21 @@ get_path_for_file (const char *folder, const char *name, CameraFileType type, Ca
 {
 	unsigned int i, l;
 	char *s, b[1024];
-	time_t t;
+	time_t t = 0;
 	struct tm *tm;
 	int hour12;
 	static int filenr = 1;
 
-	if (!file || !path)
+	if (!path)
 		return (GP_ERROR_BAD_PARAMETERS);
 
 	*path = NULL;
-	CR (gp_file_get_mtime (file, &t));
+	if (file) {
+		CR (gp_file_get_mtime (file, &t));
 	
-	if (!t)	/* use the current time as fallback if the camera did not return it. */
-		t = time(NULL);
+		if (!t)	/* use the current time as fallback if the camera did not return it. */
+			t = time(NULL);
+	}
 
 	tm = localtime (&t);
 	hour12 = tm->tm_hour % 12;
@@ -173,8 +175,17 @@ get_path_for_file (const char *folder, const char *name, CameraFileType type, Ca
 	 * If the user didn't specify a filename, use the original name 
 	 * (and prefix).
 	 */
-	if (!gp_params.filename || !strcmp (gp_params.filename, ""))
-		return gp_file_get_name_by_type (file, name, type, path);
+	if (!gp_params.filename || !strcmp (gp_params.filename, "")) {
+		if (file) {
+			return gp_file_get_name_by_type (file, name, type, path);
+		} else if ((type == GP_FILE_TYPE_NORMAL) && strchr(name,'.')) {
+			/* return the original name */
+			*path = strdup(name);
+			return (GP_OK);
+		} else
+			/* Download required to get the path from CameraFile *file */
+			return (GP_ERROR_BAD_PARAMETERS);
+	}
 
 	/* The user did specify a filename. Use it. */
 	b[sizeof (b) - 1] = '\0';
@@ -275,6 +286,18 @@ get_path_for_file (const char *folder, const char *name, CameraFileType type, Ca
 				}
 				break;
 
+			case 'F':
+				/* Get the folder name */
+				{
+					const char *f = folder ? folder : "";
+					if (f[0] == '/') f++; /* Skip first '/' */
+					if (!f[0]) f = "."; /* replace empty folder name by '.' */
+
+					strncpy (b, f, sizeof (b) - 1 - strlen(b));
+					b[sizeof(b) - 1] = '\0';
+				}
+				break;
+
 			case 'a':
 			case 'A':
 			case 'b':
@@ -292,6 +315,8 @@ get_path_for_file (const char *folder, const char *name, CameraFileType type, Ca
 			case 'Y':
 				{
 					char fmt[3] = { '%', '\0', '\0' };
+					if (!file) return (GP_ERROR_BAD_PARAMETERS); /* mtime unknown */
+
 					fmt[1] = gp_params.filename[i]; /* the letter of this 'case' */
 					strftime(b, sizeof (b), fmt, tm);
 					break;
@@ -352,6 +377,15 @@ save_camera_file_to_file (
 	free (path);
 	path = NULL;
 
+        if ((gp_params.flags & FLAGS_SKIP_EXISTING) && gp_system_is_file (s)) {
+		if ((gp_params.flags & FLAGS_QUIET) == 0) {
+			printf (_("Skip existing file %s\n"), s);
+			fflush (stdout);
+		}
+		if (curname)
+			unlink (curname);
+		return (GP_OK);
+	}
         if ((gp_params.flags & FLAGS_QUIET) == 0) {
                 while ((gp_params.flags & FLAGS_FORCE_OVERWRITE) == 0 &&
 		       gp_system_is_file (s)) {
@@ -515,6 +549,22 @@ save_file_to_file (Camera *camera, GPContext *context, Flags flags,
 	char	tmpname[20], *tmpfilename;
 	struct privstr *ps = NULL;
 
+	if (flags & FLAGS_SKIP_EXISTING && !(flags & FLAGS_STDOUT)) {
+		char *path = NULL;
+		/* Check if the file is present before downloading it. */
+		res = get_path_for_file (folder, filename, type, NULL, &path);
+
+		if (res == (GP_OK) && gp_system_is_file (path)) {
+			/* File name pattern do not require CameraFile and target file
+			   exists: Skip this file. */
+			if ((gp_params.flags & FLAGS_QUIET) == 0) {
+				printf (_("Skip existing file %s\n"), path);
+				fflush (stdout);
+			}
+			return (GP_OK);
+		}
+	}
+
 	if (flags & FLAGS_NEW) {
 		CameraFileInfo info;
 		
@@ -941,6 +991,7 @@ capture_generic (CameraCaptureType type, const char __unused__ *name, int downlo
 					(result == GP_ERROR_NO_MEMORY)		||
 					(result == GP_ERROR_CANCEL)		||
 					(result == GP_ERROR_NO_SPACE)		||
+					(result == GP_ERROR_IO_USB_CLAIM)	||
 					(result == GP_ERROR_OS_FAILURE)
 				)
 					return (result);
@@ -1024,15 +1075,20 @@ capture_generic (CameraCaptureType type, const char __unused__ *name, int downlo
 	 * so make sure we wait a bit for the the camera to finish stuff.
 	 */
 	gettimeofday (&expose_end_time, NULL);
-	waittime = 100;
-	if (glob_frames || end_next || !glob_interval || glob_bulblength) waittime = 2000;
+	waittime = 2000;
+	/* tricky, this loop might need to download both JPG and RAW. so wait longer. */
+	/*if (glob_frames || end_next || !glob_interval || glob_bulblength) waittime = 2000;*/
 	/* Drain the event queue at the end and download left over added images */
-	while ((-timediff_now(&expose_end_time)) < waittime) {
+	while (1) {
+		int realwait = waittime - (-timediff_now(&expose_end_time));
+		if (realwait < 0) realwait = 0; /* just drain the queue now */
 		result = wait_and_handle_event(waittime - (-timediff_now(&expose_end_time)), &evtype, download);
 		if ((result != GP_OK) || (evtype == GP_EVENT_TIMEOUT))
 			break;
 		if (evtype == GP_EVENT_CAPTURE_COMPLETE)
 			waittime = 100;
+		if (evtype == GP_EVENT_FILE_ADDED) /* Restart timer, more image data might come. */
+			gettimeofday (&expose_end_time, NULL);
 	}
 
 	signal(SIGUSR1, SIG_DFL);
@@ -1226,6 +1282,7 @@ typedef enum {
 	ARG_SHELL,
 	ARG_SHOW_EXIF,
 	ARG_SHOW_INFO,
+	ARG_SKIP_EXISTING,
 	ARG_SPEED,
 	ARG_STDOUT,
 	ARG_STDOUT_SIZE,
@@ -1369,6 +1426,11 @@ cb_arg_init (poptContext __unused__ ctx,
 			"option ('%s')...", arg);
 		params->p.r = action_camera_set_port (&gp_params, arg);
 		break;
+
+	case ARG_SKIP_EXISTING:
+		gp_params.flags |= FLAGS_SKIP_EXISTING;
+		break;
+
 	case ARG_SPEED:
 		params->p.r = action_camera_set_speed (&gp_params, atoi (arg));
 		break;
@@ -1828,7 +1890,7 @@ main (int argc, char **argv, char **envp)
 {
 	CallbackParams cb_params;
 	poptContext ctx;
-	int help_option_given = 0;
+	int i, help_option_given = 0;
 	int usage_option_given = 0;
 	char *debug_logfile_name = NULL;
 	const struct poptOption generalOptions[] = {
@@ -1988,6 +2050,8 @@ main (int argc, char **argv, char **envp)
 		 N_("Process new files only"), NULL},
 		{"force-overwrite", '\0', POPT_ARG_NONE, NULL,
 		 ARG_FORCE_OVERWRITE, N_("Overwrite files without asking"), NULL},
+		{"skip-existing", '\0', POPT_ARG_NONE, NULL,
+		 ARG_SKIP_EXISTING, N_("Skip existing files"), NULL},
 		POPT_TABLEEND
 	};
 	const struct poptOption miscOptions[] = {
@@ -2094,6 +2158,11 @@ main (int argc, char **argv, char **envp)
 		CR_MAIN (debug_action (&gp_params, debug_logfile_name));
 	}
 
+	gp_log (GP_LOG_DEBUG, "main", "invoked with following arguments:");
+	for (i=1;i<argc;i++)
+		gp_log (GP_LOG_DEBUG, "main", "  %s", argv[i]);
+
+
 	/* Initialize. */
 #ifdef HAVE_PTHREAD
 	gp_camera_set_timeout_funcs (gp_params.camera, start_timeout_func,
